# -*- coding: utf-8 -*-

"""Top-level package for Music Downloader Telegram Bot."""

# version as tuple for simple comparisons
VERSION = (0, 9, 17)

__author__ = """George Pchelkin"""
__email__ = 'george@pchelk.in'
# string created from tuple to avoid inconsistency
__version__ = ".".join([str(x) for x in VERSION])
